import asyncio
import os
import ssl
from typing import Any, Optional  # noqa: F401

from gunicorn.workers.base import Worker

from .serving import Server


class GunicornWorker(Worker):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.loop: Optional[asyncio.AbstractEventLoop] = None
        if self.cfg.is_ssl:
            self.ssl_context = ssl.SSLContext(self.cfg.ssl_version)
            self.ssl_context.load_cert_chain(self.cfg.certfile, self.cfg.keyfile)
            if self.cfg.ca_certs:
                self.ssl_context.load_verify_locations(self.cfg.ca_certs)
            if self.cfg.ciphers:
                self.ssl_context.set_ciphers(self.cfg.ciphers)
            self.ssl_context.set_alpn_protocols(['h2', 'http/1.1'])
        else:
            self.ssl_context = None

    def init_process(self) -> None:
        asyncio.get_event_loop().close()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        super().init_process()

    def run(self) -> None:
        create_server = asyncio.ensure_future(self._run(), loop=self.loop)  # type: ignore

        try:
            self.loop.run_until_complete(create_server)

            self.loop.run_until_complete(self._check_alive())
        finally:
            self.loop.close()

    async def _run(self) -> None:
        for sock in self.sockets:
            await self.loop.create_server(
                lambda: Server(self.wsgi, self.loop), sock=sock, ssl=self.ssl_context,
            )

    async def _check_alive(self) -> None:
        # If our parent changed then we shut down.
        pid = os.getpid()
        try:
            while self.alive:  # type: ignore
                self.notify()

                if pid == os.getpid() and self.ppid != os.getppid():
                    self.alive = False
                    self.log.info("Parent changed, shutting down: %s", self)
                else:
                    await asyncio.sleep(1.0, loop=self.loop)
        except (Exception, BaseException, GeneratorExit, KeyboardInterrupt):
            pass


class GunicornUVLoopWorker(GunicornWorker):

    def init_process(self) -> None:
        import uvloop

        asyncio.get_event_loop().close()
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

        super().init_process()
